typedef RegroupementSpecialite = {
	var libelle : String;
	var groupes : Array<Groupe>;
}

typedef SolutionCombinaison = {
	var groupes : Array<SolutionGroupe>;
	var score : Int;
}

typedef SolutionGroupe = {
	var groupe : Groupe;
	var score : Int;
}

class Promotion {
	var estTerminale : Bool;
	public var eleves : Array<Eleve>;
	public var specialites : Array<Specialite>;
	public var alignements : Array<Array<RegroupementSpecialite>>;
	public var affectesSecondChoix : Array<Eleve>;
	public var nonAffectes : Array<Eleve>;

	static var combinaisonsAlignementsPremiere : Array<Array<Int>> = [ [ 0, 1, 2 ], [ 0, 2, 1 ], [ 1, 2, 0 ], [1, 0, 2], [ 2, 0, 1 ], [ 2, 1, 0 ] ];
	static var combinaisonsAlignementsTerminale : Array<Array<Int>> = [ [ 0, 1 ], [ 1, 0 ] ];
	var combinaisonsAlignements : Array<Array<Int>>;

	public function new(eleves : Array<Eleve>, specialites : Array<Specialite>, estTerminale : Bool) {
		this.eleves = new Array();
		for (elv in eleves) { //deep copy, except voeuxSaisis
			this.eleves.push(new Eleve(elv.nom, elv.prenom, elv.voeuxSaisis));
		}
		this.specialites = new Array();
		if (specialites != null) {
			for (spe in specialites) { //deep copy
				this.specialites.push(new Specialite(spe.libelle, spe.effectifClasse, spe.effectifReduit, spe.maxGroupes));
			}
		}
		this.estTerminale = estTerminale;
		initialiser();
	}

	function initialiser() {
		var priorite = eleves.length;
		for (eleve in eleves) {
			eleve.priorite = priorite;
			priorite--;
		}
		denombrerVoeux(); //dénombrer pour chaque spécialité le nombre d'élèves l'ayant demandé
		priorite = this.specialites.length;
		for (spe in this.specialites) { //calculer pour chaque spécialité le nombre de groupes classe et réduits
			spe.creerGroupes(estTerminale);
			spe.priorite = priorite;
			priorite--;
		}
	}

	function denombrerVoeux() {
		for (spe in this.specialites) {
			spe.effectif = 0;
		}
		for (eleve in eleves) { //pour chaque élève
			for (voeu in eleve.voeux) { //pour chaque voeu
				var found = false;
				for (spe in specialites) { //on incrémente le nombre d'élève ayant demandé la spécialité
					if (spe.libelle == voeu) {
						spe.effectif++;
						found = true;
						break;
					}
				}
				if (!found) {
					var spe = new Specialite(voeu);
					spe.effectif++;
					specialites.push(spe);
				}
			}
		}
	}

/* ==================================================================================================== */
//CREATION DES ALIGNEMENTS ET GROUPES
	public function creerAlignements() {
		var nbAlignements = estTerminale ? 2 : 3;
		alignements = new Array();

		//préparation des alignements : [ [svt[], pc[], ...], [svt[], pc[], ...], ...]
		for (i in 0...nbAlignements) {
			var ali = new Array();
			for (spe in specialites) {
				ali.push({ libelle : spe.libelle, groupes : new Array() });
			}
			alignements.push(ali);
		}

		//création des groupes
		var ali : Int = 0;
		for (spe in specialites) {
			var i : Int = 0;
			while (i < spe.nbGroupesClasse) { //on créer d'abord les groupes classe
				var regspe = getRegroupementSpecialite(ali, spe.libelle);
				var libelle = spe.libelle + "-" + (i + 1);
				regspe.groupes.push(new Groupe(libelle, spe.effectifClasse));
				ali = (ali + 1) % nbAlignements; //rotation de l'alignement pour répartir les groupes de la spécialité dans les différents alignements
				i++;
			}
			while (i < spe.nbGroupesClasse + spe.nbGroupesReduits) { //puis les groupes réduits
				var regspe = getRegroupementSpecialite(ali, spe.libelle);
				var libelle = spe.libelle + "-" + (i + 1);
				regspe.groupes.push(new Groupe(libelle, spe.effectifReduit));
				ali = (ali + 1) % nbAlignements; //rotation de l'alignement pour répartir les groupes de la spécialité dans les différents alignements
				i++;
			}
		}
	}

/* ==================================================================================================== */
//FONCTIONS UTILITAIRES
	function getSpecialite(libelle : String) : Specialite {
		var specialite : Specialite = null;
		for (spe in specialites) {
			if (spe.libelle == libelle) {
				specialite = spe;
				break;
			}
		}
		return specialite;
	}

	function getRegroupementSpecialite(ali : Int, libelle : String) : RegroupementSpecialite {
		var regroupementSpecialite : RegroupementSpecialite = null;
		for (regspe in alignements[ali]) { //recherche de la spécialité dans l'alignement
			if (regspe.libelle == libelle) {
				regroupementSpecialite = regspe;
				break;
			}
		}
		return regroupementSpecialite;
	}

/* ==================================================================================================== */
//TRI / OPTIMISATION
	public function trierSpecialites(croissant = false) {
		for (spe in specialites) {
			var nbGroupes = spe.nbGroupesClasse + spe.nbGroupesReduits;
			var placesDispos = spe.capacite - spe.effectif;
			if (spe.capacite > 0) {
				//la priorité est inversement proportionnelle au nombre de groupes et de places inutilisées:
				spe.priorite = Math.round(1000 / (nbGroupes * 10 + placesDispos / nbGroupes));
			}
		}
		specialites.sort(function(s1, s2) {
			return croissant ? 1 : -1 * s1.priorite == s2.priorite ? 0 : s1.priorite < s2.priorite ? 1 : -1;
		});
	}

	public function trierEleves(croissant = false) {
		for (eleve in eleves) {
			for (voeu in eleve.voeux) {
				eleve.priorite += getSpecialite(voeu).priorite;
			}
		}
		eleves.sort(function(e1, e2) {
			return croissant ? 1 : -1 * e1.priorite == e2.priorite ? 0 : e1.priorite < e2.priorite ? 1 : -1;
		});
	}

/* ==================================================================================================== */
//AFFECTATION DES ELEVES
	public function affecterEleves() {
		nonAffectes = new Array();
		affectesSecondChoix = new Array();
		combinaisonsAlignements = estTerminale ? combinaisonsAlignementsTerminale : combinaisonsAlignementsPremiere;
		for (eleve in eleves) {
			var solCmbAli = rechercheCombinaison(eleve); //recherche d'une combinaison d'alignements permettant d'affecter l'élève
			if (solCmbAli != null) { //on affecte l'élève dans chaque groupe
				for (solgrp in solCmbAli.groupes) {
					solgrp.groupe.elevesAffectes.push(eleve);
				}
			} else if (eleve.voeuxSaisis.length == 4 && getSpecialite(eleve.voeuxSaisis[3]) != null) { //on essaie le second choix
				eleve.secondChoix();
				solCmbAli = rechercheCombinaison(eleve);
				eleve.restaurerChoix();
				if (solCmbAli != null) {
					affectesSecondChoix.push(eleve);
					for (solgrp in solCmbAli.groupes) {
						solgrp.groupe.elevesAffectes.push(eleve);
					}
				} else {
					nonAffectes.push(eleve);
				}
			} else {
				nonAffectes.push(eleve);
			}
		}
	}

	function rechercheCombinaison(eleve : Eleve) : SolutionCombinaison {
		var meilleureSolCmb : SolutionCombinaison = null;
		var meilleurScore : Int = 0;
		var meilleurMini : Int = 0;
		for (cmbali in combinaisonsAlignements) { //on teste chaque combinaison d'alignement
			var okcmb = true; //on considère par défaut que la combinaison permet d'affecter les élèves
			var solCmbAli : SolutionCombinaison = { groupes : new Array(), score : 0 };
			for (i in 0...eleve.voeux.length) { //pour chaque spécialité, on regarde s'il y a de la place dans au moins un groupe de cet alignement
				var regspe = getRegroupementSpecialite(cmbali[i], eleve.voeux[i]);
				var groupe = rechercheGroupe(regspe); //recherche d'un groupe (dans la spécialité) pouvant accueillir l'élève
				if (groupe == null) {
					okcmb = false;
					break;
				} else {
					solCmbAli.groupes.push(groupe);
					solCmbAli.score += groupe.score;
				}
			}
			if (okcmb) {
				var mini : Int = 120; //40x3
				for (grp in solCmbAli.groupes) {
					if (grp.score < mini) {
						mini = grp.score;
					}
				}
				if (mini > meilleurMini || mini == meilleurMini && solCmbAli.score > meilleurScore) {
					meilleurScore = solCmbAli.score;
					meilleureSolCmb = solCmbAli;
					meilleurMini = mini;
				}
			}
		}
		return meilleureSolCmb;
	}

	function rechercheGroupe(regspe : RegroupementSpecialite) : SolutionGroupe {
		var meilleureSolGrp : SolutionGroupe = null;
		var meilleureDispo = 0;
		var cumulDispo = 0; //cumul des disponibilites de l'ensemble des groupes
		var meilleurGroupe : Groupe = null;
		for (grp in regspe.groupes) {
			var dispo = grp.capacite - grp.elevesAffectes.length;
			cumulDispo += dispo;
			if (dispo > meilleureDispo) {
				meilleureDispo = dispo;
				meilleurGroupe = grp;
			}
		}
		if (cumulDispo > 0) {
			meilleureSolGrp = { groupe : meilleurGroupe, score : cumulDispo };
		}
		return meilleureSolGrp;
	}

/* ==================================================================================================== */
//STATISTIQUES ET CONTROLE
	public function getEffectifsSpecialites() : Int {
		var nbEleves = 0;
		for (spe in specialites) {
			nbEleves += spe.effectif;
		}
		return nbEleves;
	}
}
